/* MPFS's TV background cross center detection */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/times.h>

#include "tv_image_lib.h"

typedef int bool;
#define TRUE    ((bool)1)
#define FALSE   ((bool)0)

bool get_fits(char *fitsname, int col0, int lin0, int frmw, int frmh, short *array, int *tv_lev0, int *tv_elref, int *nadd);

void fft1c(double *re_d, double *im_d, int n, int sign);

double mxmn(double f_re[128], int length);

void main(int argc, char *argv[]) {
   short *img;                     /* image memory */
   int imgw, imgh;                 /* image size   */
   int nchan = 0;                  /* TV-chan. number */
   short *frm1=NULL,*frm2=NULL;    /* frames memory */
   int col0, lin0, frmw, frmh;     /* frame1: area of interest */
   int colk, link;
   int coln, linn, colc, linc;
   int bg_lev0,bg_nadd;
   int col20, lin20;               /* frame2 */
   int width, height;              /* FFT frame size*/
   int nseq;
   int i,j,k,l,x,y,pix1,pix2;
   int n,nx,ny;
   double fx_re[128],  fx_im[128];
   double fx1_re[128], fx1_im[128];
   double fx2_re[128], fx2_im[128];
   double fy_re[128],  fy_im[128];
   double fy1_re[128], fy1_im[128];
   double fy2_re[128], fy2_im[128];
   double xfilt[128],  yfilt[128];
   double s1, s2, xmx=0., ymx=0.;
   double a11, a12, a21, a22, b1, b2, a, b;
   char tmp[100],fname[100];
   char *bgdfile=NULL;

   if (argc>1) nchan = atoi(argv[1]);
   if (argc>2) bgdfile = argv[2];

   get_vs56_sh_mem();
   img = tv_chan_image (nchan);
   imgw = tv_chan_img_width();
   imgh = tv_chan_img_height();
   nseq = tv_chan_seq_num();

reload_bgd:
   if(isatty(0)) {
      fputs("Define the 1-st frame ...",stdout);
      fgets(tmp,10,stdin);
   }
   tv_chan_get_frame(&col0, &lin0, &frmw, &frmh);

   width=frmw; height=frmh;

   if(col0==0 && lin0==0 && frmw==imgw && frmh==imgh ) {
      fprintf(stderr,"WARNING! No Frame defined for chan.%d\n",nchan);
   }

   if(frm1!=NULL) free(frm1);
   frm1 = (short *)malloc(frmw*frmh*sizeof(short));

   if(bgdfile!=NULL) {
      if(!get_fits(bgdfile, col0,lin0,frmw,frmh, frm1, &bg_lev0, &i, &bg_nadd)) {
	 fputs("quit\n",stdout);
	 fflush(stdout);
	 exit(1);
      }
   } else {
      for (j=0; j<frmh; j++) {
	 y=lin0+j;
	 for (i=0; i<frmw; i++) {
	    x=col0+i;
	    frm1[frmw*j+i] = img[imgw*y+x];
	 }
      }
      fprintf(stderr, "Load BGD-frame from TV-image: %d,%d %d x %d\n",
	      col0,lin0,frmw,frmh);
   }
get_command:
   if(isatty(0))
      fputs("Point to the 2-nd one ... (or type 'q' to quit) ",stdout);
   fgets(tmp,99,stdin);
   if(tmp[0]=='q')      exit(0);           /* quit */
   else if(tmp[0]=='r') goto reload_bgd;   /* reload */
   else if(tmp[0]=='f') {                  /* file .... */
       char *p;
       p=strtok(tmp," =\t\n");
       p=strtok(NULL," =\t\n");
       if(p!=NULL) {
	  if(strcmp(p,"no")) {
	     strncpy(fname,p,90);          /* file fname.mt */
	     bgdfile=fname;
	  } else
	     bgdfile=NULL;                 /* file no */
       } else
	  bgdfile=NULL;                    /* file */
       goto get_command;
   } else if(tmp[0]=='n' || tmp[0]=='\n')  /* next */
	goto next_frame;
   else if(tmp[0]=='s')                    /* subtract */
	goto subtruct_frm;
   else if(tmp[0]=='z') {                  /* zero ...*/
	if(strchr(tmp,'x'))
	   xmx = 0.;                       /* zero x */
	if(strchr(tmp,'y'))
	   ymx = 0.;                       /* zero y */
	if(strchr(tmp,'x')==NULL && strchr(tmp,'y')==NULL)
	   xmx = ymx = 0.;                 /* zero */
	goto get_command;
   } else goto get_command;

next_frame:
   tv_chan_get_frame(&col20, &lin20, &i, &j);

   if(frm2!=NULL) free(frm2);
   frm2 = (short *)malloc(frmw*frmh*sizeof(short));

   for (j=0; j<frmh; j++) {
      y=lin20+j;
      for (i=0; i<frmw; i++) {
	 x=col20+i;
	 frm2[frmw*j+i] = img[imgw*y+x];
      }
   }

   coln = col0+frmw;
   colc = col0+frmw/2;
   for(nx=0; width!=0; width >>= 1, nx++);
   nx -= 1;
   if(nx<2) nx=2;
   if(nx>7) nx=7;
   width = 1<<nx;

   linn = lin0+frmh;
   linc = lin0+frmh/2;
   for(ny=0; height!=0; height >>= 1, ny++);
   ny -= 1;
   if(ny<2) ny=2;
   if(ny>7) ny=7;
   height = 1<<ny;

   n = width*height;
   for (i=0; i<width; i++) {
      fy1_re[i]=fy2_re[i]=0.;
      fy1_im[i]=fy2_im[i]=0.;
   }
   for (i=0; i<height; i++) {
      fx1_re[i]=fx2_re[i]=0.;
      fx1_im[i]=fx2_im[i]=0.;
   }
   xfilt[0]=1.;
   for (i=1; i<width/2; i++)
      xfilt[i]=xfilt[width-i]=(i<width/4)? 1.0 : (i<width/3)? ((double)width/3.-i)/(width/12.) : 0.0;
   /*for(i=0; i<width; i++) printf("%.2f ",xfilt[i]);printf("\n");*/
   yfilt[0]=1.;
   for (i=1; i<height/2; i++)
      yfilt[i]=yfilt[height-i]=(i<height/4)? 1.0 : (i<height/3)? ((double)height/3.-i)/(height/12.) : 0.0;
   /*for(i=0; i<height; i++) printf("%.2f ",yfilt[i]);printf("\n");*/

   a11=a12=a21=a22=b1=b2=0.0;

    colk=frmw/2-width/2;
    s1=s2=0.;
    for (j=0; j<frmh; j++)
       if(j<frmh/3 || j>frmh*2/3) {
	  for (i=0; i<width; i++) {
	     k = frmw*j+colk+i;
	     pix1 = frm1[k];
	     fy1_re[i] += pix1;
	     s1 += pix1;
	     pix2 = frm2[k];
	     fy2_re[i] += pix2;
	     s2 += pix2;
	     a11 += pix1*pix1;
	     a22 += 1.;
	     b1 += pix1*pix2;
	  }
       }
    a12 = s1;
    a21 = a12;
    b2 = s2;

    fft1c(fy1_re, fy1_im, width, 1);
    fft1c(fy2_re, fy2_im, width, 1);

    for (i=0; i<width; i++) {
       double flt,r0,r1,r2,i0,i1,i2;
       flt=xfilt[i];
       r1=fy1_re[i]/s1;
       r2=fy2_re[i]/s2;
       i1=fy1_im[i]/s1;
       i2=fy2_im[i]/s2;
       r0 = (r1*r2+i1*i2)*flt;
       i0 = (r1*i2-r2*i1)*flt;
       fy_re[i] = r0;
       fy_im[i] = i0;
    }
    fft1c(fy_re, fy_im, width, -1);

    link=frmh/2-height/2;
    s1=s2=0.;
    for (j=0; j<frmw; j++)
       if(j<frmw/3 || j>frmw*2/3) {
	  for (i=0; i<height; i++) {
	     k = frmw*(link+i)+j;
	     pix1 = frm1[k];
	     fx1_re[i] += pix1;
	     s1 += pix1;
	     pix2 = frm2[k];
	     fx2_re[i] += pix2;
	     s2 += pix2;
	     a11 += pix1*pix1;
	     a22 += 1.;
	     b1 += pix1*pix2;
	  }
      }
    a12 += s1;
    a21 = a12;
    b2 += s2;
    b=(a11*b2-a21*b1)/(a11*a22-a21*a12);
    a=(b1-b*a12)/a11;
/*fprintf(stderr,"A=%.5f B=%.5f Avr1=%.5f Avr2=%.5f \n",a,b,a12/a22,b2/a22);*/

    fft1c(fx1_re, fx1_im, height, 1);
    fft1c(fx2_re, fx2_im, height, 1);

    for (i=0; i<height; i++) {
       double flt,r0,r1,r2,i0,i1,i2;
       flt=yfilt[i];
       r1=fx1_re[i]/s1;
       r2=fx2_re[i]/s2;
       i1=fx1_im[i]/s1;
       i2=fx2_im[i]/s2;
       r0 = (r1*r2+i1*i2)*flt;
       i0 = (r1*i2-r2*i1)*flt;
       fx_re[i] = r0;
       fx_im[i] = i0;
    }
    fft1c(fx_re, fx_im, height, -1);

/*for (i=0; i<height; i++) printf("%d %.5f %.5f %.5f %.5f\n",i,fy_re[i],fy_im[i],fx_re[i],fx_im[i]);*/

   ymx = mxmn(fx_re, height);
   xmx = mxmn(fy_re, width);
   printf("dX=%.2f\ndY=%.2f\n",xmx,ymx);
   fflush(stdout);
   goto get_command;

subtruct_frm:
   {
    int ix= -(int)xmx;
    double dx= -xmx-ix;
    int iy= -(int)ymx;
    double dy= -ymx-iy;
    if(dy<0.) {iy--; dy += 1.0;}
    if(dx<0.) {ix--; dx += 1.0;}

    tv_chan_get_frame(&col20, &lin20, &i, &j);

    if(frm2!=NULL) free(frm2);
    frm2 = (short *)malloc(frmw*frmh*sizeof(short));

    a11=a12=a21=a22=b1=b2=0.0;
/*fprintf(stderr,"ix=%d,dx=%.5f iy=%d dy=%.5f\n",ix,dx,iy,dy);fflush(stderr);*/
    for (j=0; j<frmh; j++) {
      int j0,j1;
      j0=j+iy;
      j1=j0+1;
      if(j0<0) j0 = -j0;
      if(j0>=frmh) j0 = 2*frmh-j0-1;
      if(j1<0) j1 = -j1;
      if(j1>=frmh) j1 = 2*frmh-j1-1;
      y=lin20+j;
      for (i=0; i<frmw; i++) {
	 int i0,i1,p00,p01,p10,p11;
	 i0=i+ix;
	 i1=i0+1;
	 if(i0<0) i0 = -i0;
	 if(i0>=frmw) i0 = 2*frmw-i0-1;
	 if(i1<0) i1 = -i1;
	 if(i1>=frmw) i1 = 2*frmw-i1-1;
	 x=col20+i;
	 k = imgw*y+x;
	 l = frmw*j+i;
	 p00 = frm1[frmw*j0+i0];
	 p01 = frm1[frmw*j0+i1];
	 p10 = frm1[frmw*j1+i0];
	 p11 = frm1[frmw*j1+i1];
	 pix1 = (p00*(1.-dx)+p01*dx)*(1.-dy) + (p10*(1.-dx)+p11*dx)*dy;
	 pix2 = img[k];
	 frm2[l] = pix1;
	 if(j<frmh/3 || j>frmh*2/3 || i<frmw/3 || i>frmw*2/3) {
	    a11 += pix1*pix1;
	    a12 += pix1;
	    a22 += 1.;
	    b1 += pix1*pix2;
	    b2 += pix2;
	 }
      }
    }
    a21 = a12;
    b=(a11*b2-a21*b1)/(a11*a22-a21*a12);
    a=(b1-b*a12)/a11;
/*fprintf(stderr,"A=%.5f B=%.5f Avr1=%.5f Avr2=%.5f \n",a,b,a12/a22,b2/a22);*/
    for (j=0; j<frmh; j++) {
      y=lin20+j;
      for (i=0; i<frmw; i++) {
	 x=col20+i;
	 k = imgw*y+x;
	 l = frmw*j+i;
	 pix1 = frm2[l];
	 pix2 = img[k];
	 img[k] = (short)(pix2 - pix1*a - b + 0.5);
      }
    }
   }
   tv_chan_set_seq_num(tv_chan_seq_num()+1);
   goto get_command;
}

double mxmn(double f_re[128], int length) {
   int i,imx,imn;
   double fmx,fmn,xmx;
   int i0,i1,i2,i3;
   double f0,f1,f2,f3;

   fmx=0.;
   imx=0;
   for (i=0; i<length; i++) {
      double f;
      f = f_re[i];
      if (f>fmx) {fmx=f;imx=i;}
   }

   i0=imx-1;if(i0<0) i0+=length;
   i1=imx;
   i2=imx+1;if(i2>=length) i2-=length;
   if(f_re[i0] > f_re[i2] ) {
      i0=imx-2; if(i0<0) i0+=length;
      i1=imx-1; if(i1<0) i1+=length;
      i2=imx;
   }
   i3=i2+1;if(i3>=length) i3-=length;
   f0 = f_re[i0]; f1 = f_re[i1];
   f2 = f_re[i2]; f3 = f_re[i3];
   xmx = i0 + 1.5 - (f2-f1)/((f0+f3-f1-f2)/2.0);
/*printf("%.4f\t%.4f\t%.4f\t%.4f\n",f0,f1,f2,f3);*/
/*printf("%d\t%d\t%d\t%d\n",i0,i1,i2,i3);*/
   fmn=fmx;
   imn=0;
   for (i=0; i<length; i++) {
      double f;
      f = f_re[i];
      if (f<fmn) {fmn=f;imn=i;}
   }
   if(imx>length/2) imx -= length;
   if(xmx>length/2) xmx -= length;
   if(imn>length/2) imn -= length;
/*fprintf(stderr,"Max:%.5f(%.2f) Min:%.5f(%d) \n",fmx,xmx,fmn,imn);*/

   return(xmx);
}
