#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Load frame from FITS-file */
typedef int bool;
#define TRUE    ((bool)1)
#define FALSE   ((bool)0)

bool get_fits(char *fitsname, int col0, int lin0, int frmw, int frmh, short *array, int *tv_lev0, int *tv_elref, int *nadd)
{
  int k,l,x,y,lin;
  char tmp[2880];
  FILE *fd;
  bool simple = FALSE;
  int bitpix = 8;
  int naxis = 0;
  int xsize = 0;
  int ysize = 0;
  double bscale = 1.;
  double bzero = 0.;
  int mn=32767, mx=-32768;
  bool pipe = FALSE;
  short *lbuf = NULL;

  if(strstr(fitsname,".gz")) {
     char command[100];
     sprintf(command,"zcat %s",fitsname);
     if((fd = popen(command,"r")) == NULL) {
	fprintf(stderr, "Can't unZIP FITS-file %s\n", fitsname);
	return FALSE;
     }
     pipe = TRUE;
  } else {
     if((fd=fopen(fitsname, "r")) == NULL) {
	fprintf(stderr, "Can't read FITS-file %s\n", fitsname);
	return FALSE;
     }
     pipe = FALSE;
  }
  for (lin=36*5; lin>0; lin--) {
     char *pname = tmp;
     char *pval = tmp+9;
     char *pcom ;
     if (fread(tmp,1,80,fd) <= 0 ) goto wrong_fits;
     pcom = strchr(pval, '/');
     if(pcom != NULL) *pcom++ = '\0';

     if(       strncmp(pname,"END      ",9)==0) break;
     else if(  strncmp(pname,"SIMPLE  =",9)==0) {
	if( strstr(pval," T") == NULL ) goto wrong_fits;
	simple = TRUE;
     } else if(strncmp(pname,"BITPIX  =",9)==0) {
	if( strstr(pval," 16")  == NULL) goto wrong_fits;
	bitpix = 16;
     } else if(strncmp(pname,"NAXIS   =",9)==0) {
	if( strstr(pval," 2") == NULL) goto wrong_fits;
	naxis = 2;
     } else if(strncmp(pname,"NAXIS1  =",9)==0) {
	xsize = atoi(pval);
     } else if(strncmp(pname,"NAXIS2  =",9)==0) {
	ysize = atoi(pval);
     } else if(strncmp(pname,"BSCALE  =",9)==0) {
	bscale = atof(pval);
     } else if(strncmp(pname,"BZERO   =",9)==0) {
	bzero = atof(pval);
     } else if(strncmp(pname,"TV-BZERO=",9)==0) {
	*tv_lev0 = atoi(pval);
     } else if(strncmp(pname,"TV-ELREF=",9)==0) {
	*tv_elref = atoi(pval);
     } else if(strncmp(pname,"TV-NADD =",9)==0) {
	*nadd = atoi(pval);
     }
  }
/*printf("%dlines simple=%s bitpix=%d naxis=%d\n",36*5-lin,simple?"T":"F",bitpix,naxis);*/
/*printf("xsize=%d  ysize=%d bscale=%f bzero=%f\n",xsize,ysize,bscale,bzero);*/

  if(lin==0 || !simple || bitpix!=16 || naxis!=2 || xsize<1 || ysize<1) goto wrong_fits;

  lbuf = (short *)calloc(xsize, 2);
  if(lbuf == NULL) {
	fprintf(stderr, "Can't allocate %d-byte buffer\n", xsize*2);
	return FALSE;
  }
  if(pipe) {
     lin = lin%36;
     while(--lin>0)
	if (fread(tmp,1,80,fd) <= 0 ) goto wrong_fits;
     for ( y=0; y<lin0; y++ )
	if (fread(lbuf,2,xsize,fd) <= 0 ) goto wrong_fits;
  } else
     fseek(fd, ((ftell(fd)-1)/2880+1)*2880 + lin0*xsize*2, SEEK_SET);

  for ( l=0,y=lin0; y<ysize && l<frmh; y++,l++ ) {
     if(pipe) {
	if (col0>0) {
	   if (fread(lbuf,2,col0,fd) <= 0 ) goto wrong_fits;
	}
     } else
	fseek(fd, col0*2, SEEK_CUR);
     for ( k=0,x=col0; x<xsize && k<frmw; x++,k++ ) {
	union {char b[2]; short i;} pix;
	int i;
	if ((i=fgetc(fd))==EOF)  goto wrong_fits;
	pix.b[1] = i;
	if ((i=fgetc(fd))==EOF)  goto wrong_fits;
	pix.b[0] = i;
	i = array[l*frmw+k] = (short)(pix.i*bscale + bzero + 0.5);
	if(mn>i) mn = i;
	if(mx<i) mx = i;
     }
     if (x<xsize) {
	if(pipe) {
	   if (fread(lbuf,2,xsize-x,fd) <= 0 ) goto wrong_fits;
	} else
	   fseek(fd, (xsize-x)*2, SEEK_CUR);
     }
  }
  fclose(fd);
  if(lbuf!=NULL) free(lbuf);
  fprintf(stderr, "Load FITS from %s  %d x %d ",fitsname,xsize,ysize);
  fprintf(stderr, "frame: %d,%d %d x %d  min=%d max=%d\n",col0,lin0,frmw,frmh,mn,mx);
  return TRUE;

wrong_fits:
  fprintf(stderr, "Can't recognize FITS-format (%s)\n", fitsname);
  fclose(fd);
  if(lbuf!=NULL) free(lbuf);
  return FALSE;
}
