#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <time.h>

#include "tvimage.h"
#include "tv_image_lib.h"

#include <qobject.h>
#include <qpopmenu.h>
#include <qkeycode.h>
#include <qimage.h>
#include <qtimer.h>
#include <qapp.h>
#include <qlayout.h>
#include <qdir.h>
#include <qfile.h>
#include <qfiledlg.h>
#include <qtstream.h>
#include <qtooltip.h>
#include <qmlined.h>

#include "tv_image_xpm.h"


TVimage::TVimage( QWidget *parent, const char *name )
    : QWidget( parent, name)
{
    QPixmap p1( p1_xpm );
    QPixmap p2( p2_xpm );
    QPixmap p3( p3_xpm );
    QPixmap p4( p4_xpm );
    QPixmap p5( p5_xpm );
    QPixmap p6( p6_xpm );
    QPixmap p7( p7_xpm );
    QPixmap p8( p8_xpm );
    QPixmap p9( p9_xpm );
    QPixmap p10( p10_xpm );
    QPixmap p11( p11_xpm );
    QPixmap p12( p12_xpm );
    QPixmap cycl( cycle_xpm );
    QPixmap excl( exclam_xpm );
    QPixmap sdisk( save_xpm );
    QPixmap ldisk( load_xpm );
    QPixmap thelp( text_xpm );
    QPixmap nscape( nscape_xpm );

    oper = new QPopupMenu();
    CHECK_PTR( oper);
    oper->insertItem( p4, "G&et single image", this, SLOT(single()), CTRL+Key_1 );
    oper->insertItem( p5, "Get te&n images",   this, SLOT(get10()),  CTRL+Key_0 );
    oper->insertItem( p5, "&Get infinitely",   this, SLOT(cont()),   CTRL+Key_G );
    oper->insertItem( p8, "&Stop inf.mode",    this, SLOT(stop()),   CTRL+Key_S );
    oper->insertItem( p6, "&Add one",     this, SLOT(add1()),   CTRL+Key_A );
    oper->insertItem( p7, "Add &ten",     this, SLOT(add10()),  CTRL+Key_T );
    oper->insertItem( p7, "Add in&f-ly",  this, SLOT(addcont()),CTRL+Key_C );
    oper->insertItem( cycl, "Start cycling",  this, SLOT(start_cycle()),CTRL+Key_N );
    oper->insertSeparator();
    oper->insertItem( p9,  "E&rase image", this, SLOT(clear()), CTRL+Key_R );
    oper->insertItem( p12, "&Write VS56",  this, SLOT(write()), CTRL+Key_W );
    oper->insertItem( p10, "&Init VS56",   this, SLOT(init()),  CTRL+Key_I );

    file = new QPopupMenu();
    CHECK_PTR( file);
    save = new QPopupMenu();
    CHECK_PTR( save);
    save->insertItem( "Gray&Scale JPEG",  0 );
    save->insertItem( "&GrayScale GIF",   1 );
    save->insertItem( "Color &JPEG",      2 );
    save->insertItem( "&Color GIF",       3 );
    connect( save, SIGNAL(activated(int)), this, SLOT(save_to_file(int)) );

    file->insertItem( sdisk, "&Save picture",  save );
    file->insertItem( sdisk, "Save as &FITS ", this, SLOT(save_fits()),  CTRL+Key_F );
    file->insertItem( ldisk, "&Load FITS",     this, SLOT(load_fits()),  CTRL+Key_L );
    file->insertSeparator();
    file->insertItem( excl, "E&xit",  qApp, SLOT(quit()), CTRL+Key_Q );

    mmod = new QPopupMenu();
    CHECK_PTR( mmod);
    int id;
    id=mmod->insertItem( "Modify &CUTs",MOD_CUTS); mmod->setAccel(SHIFT+Key_M, id);
    id=mmod->insertItem( "Set &Frame",  SET_FRAME);mmod->setAccel(SHIFT+Key_S, id);
    id=mmod->insertItem( "Set &Line",   SET_LINE); mmod->setAccel(SHIFT+Key_G, id);
    id=mmod->insertItem( "Set &Point",  SET_POINT);mmod->setAccel(SHIFT+Key_P, id);
    mmod->insertSeparator();
    id=mmod->insertItem( "&No Mouse",   NO_MOUSE); mmod->setAccel(SHIFT+Key_N, id);
    connect( mmod, SIGNAL(activated(int)), this, SLOT(setMouseMode(int)) );

    opts = new QPopupMenu();
    CHECK_PTR( opts);
    opts->insertItem( "&Calculate CUTs",this, SLOT(stat_image()), SHIFT+Key_C);
    opts->insertItem( "Change &Format", this, SLOT(togleFormat()),SHIFT+Key_F);
    opts->insertItem( "&Reset Frame",   this, SLOT(resetFrame()), SHIFT+Key_R);
    opts->insertItem( "Set E&xposure",  this, SLOT(startExp()),   SHIFT+Key_X);
    opts->insertItem( "Set &El.Ref.Lev.", this, SLOT(startElev()),SHIFT+Key_E);
    opts->insertItem( "Load &LUT",      this, SLOT(getLUT()),     SHIFT+Key_L);
    opts->insertItem( "Load &ITT",      this, SLOT(getITT()),     SHIFT+Key_I);
    opts->insertItem( "&Display LUT",   this, SLOT(togleBar()),   SHIFT+Key_D);
    opts->insertItem( "Show &Opt.Window",this, SLOT(setup_opt()), SHIFT+Key_O);

    ccd = new QPopupMenu();
    CHECK_PTR( ccd);
    ccd->insertItem( "&Calc BgdRow", this, SLOT(calc_flat_row()), SHIFT+Key_B);
    ccd_id = ccd->insertItem("Vid ->ccd (Use &BgdRow online)", this, SLOT(togle_ccd_mode()), SHIFT+Key_V);
       QPopupMenu *frop = new QPopupMenu();
       CHECK_PTR( frop);
       id=frop->insertItem( "-",  1 ); frop->setAccel(ALT+Key_Minus, id);
       id=frop->insertItem( ":",  3 ); frop->setAccel(ALT+Key_Semicolon, id);
       id=frop->insertItem( "+",  0 ); frop->setAccel(ALT+Key_Equal,  id);
       id=frop->insertItem( "*",  2 ); frop->setAccel(ALT+Key_8,id);
       connect( frop, SIGNAL(activated(int)), this, SLOT(use_flat_row(int)) );
    ccd->insertItem("Image = Image &o BgdRow", frop );

    math = new QPopupMenu();
    CHECK_PTR( math);
    math->insertItem( "Set &BGD-file name",   this, SLOT(set_bias_name()));
    math->insertItem( "Set &FLAT-file name",  this, SLOT(set_flat_name()));
    math->insertItem( "&Subtract BGD",        this, SLOT(subtr_bias()), CTRL+Key_B );
    math->insertItem( "&Divide by FLAT",      this, SLOT(div_by_flat()), CTRL+Key_D );
    math->insertItem( "(&Image-BGD)/FLAT",    this, SLOT(bias_flat_red()), CTRL+Key_M );

    help = new QPopupMenu();
    CHECK_PTR( help);
    help->insertItem( thelp, "&Plain text",   this, SLOT(show_help()));
    help->insertItem( nscape,"&HTML manual",  this, SLOT(html_manual()));

    menu = new QMenuBar( this );
    CHECK_PTR( menu );
    menu->insertItem( "&File", file );
    menu->insertItem( "&VS56 action", oper );
    menu->insertItem( "&Mouse mode", mmod );
    menu->insertItem( "&Options", opts );
    menu->insertItem( "&CCD-spec.", ccd );
    menu->insertItem( "&Reduction", math );
    menu->insertSeparator();
    menu->insertItem( "&Help", help );
    //menu->setSeparator( QMenuBar::InWindowsStyle );

    mesg = new QLabel( this );
    CHECK_PTR( mesg );
    mesg->setGeometry( 10, menu->height(), width()-20, 20 );
    mesg->setFrameStyle( QFrame::Box | QFrame::Raised );
    mesg->setLineWidth( 1 );
    mesg->setAlignment( AlignCenter );

    image_line0 = menu->height() + mesg->height();

    nchan = 0;
    wtag = NULL;
    td_opt = NULL;
    cutTop = 200;
    cutBot = 10;
    mouse_mode = NO_MOUSE;
    mouse_button = NoButton;
    force_showing = TRUE;
    savename = NULL;
    savedir = ".";
    biasname = NULL;
    flatname = NULL;
    servdir = ".";
    show_bar = FALSE;
    ccd_mode = FALSE;
    image_w=512;
    image_h=512;
    cycle_n=1;
    cycle_stage=0;
    cycle_nadd=5;
    cycle_delay=2.5;
    cycle_bias = FALSE;
    cycle_flat = FALSE;

    lin.begin(this);
    lin.setBackgroundMode(OpaqueMode);
    QColor c;
    c.setNamedColor("Black");
    lin.setBackgroundColor(c);
    c.setNamedColor("White");
    lin.setPen(c);
    lin.setRasterOp (XorROP);
    //lin.setPen(DashLine);
    //graph.begin(this);
    //graph.setPen(c);

    image_pm.resize(image_w,image_h);

    QPainter p;
    p.begin( &image_pm );
    c.setNamedColor("Red");
    p.setPen(c);
    p.setFont( QFont( "times", 48, QFont::Bold ) );
    p.drawText(0,0,image_w,image_h, AlignCenter, "TVimage");
    p.end();

    for ( int i=0; i<256; i++ ) itt[i] = i;             // build default ITT
    getITT("ramp.itt");
    for ( int i=0; i<256; i++ ) lut[i] = qRgb(i,i,i);   // build grey color table
    getLUT("ramp.lut");

    tvtimer = new QTimer(this);              // create internal timer
    connect( tvtimer, SIGNAL(timeout()), this, SLOT(tmout()) );
    tvtimer->start( 1000, TRUE );           // emit 1-st signal after 1.0 seconds
}
TVimage::~TVimage()
{
    lin.end();
}
void TVimage::setChan( int n)
{
    nchan = n;
    tv_image = tv_chan_image (nchan);
    image_w = tv_chan_img_width();
    image_h = tv_chan_img_height();
    format  = tv_chan_form(-1);
}
void TVimage::setTagWnd( QWidget *w )
{
    wtag = w;
}

void TVimage::setup_opt()  // Show "Options" dialog panel
{
    if (td_opt != NULL) {
       if(td_opt->isVisible())
	  td_opt->raise();
       else {
	  td_opt->setGeometry( x() + width() - 300, y(), 200, 200 );
	  td_opt->show();
       }
       return;
    }
    char str[40],cap[40];
    sprintf(cap,"TVimage(%d) Options", nchan);
    td_opt = new QTabDialog( 0, "opt_dialog" );
    connect( td_opt, SIGNAL(destroyed()), SLOT(destr_opt()) );
    td_opt->setGeometry( x() + width() - 300, y(), 200, 200 );
    td_opt->setCaption( cap );

    // set up 1-st page  of the tab dialog
    QWidget * w = new QWidget( td_opt, "cycle" );
    QGridLayout * g = new QGridLayout( w, 4, 2, 5 );

    ed_c_nadd = new QLineEdit( w );
    g->addWidget( ed_c_nadd, 0, 1 );
    sprintf(str,"%d",cycle_nadd);
    ed_c_nadd->setText( str );
    ed_c_nadd->setMinimumSize( 30, 15 );
    ed_c_nadd->setMaximumSize( 100, 30 );
    QToolTip::add(ed_c_nadd, "Number of frames to ADD" );
    QLabel * lb = new QLabel( w );
    g->addWidget( lb, 0, 0 );
    lb->setText( "Add N:" );
    lb->setBuddy( ed_c_nadd );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Number of frames to ADD" );

    ed_c_dt = new QLineEdit( w );
    g->addWidget( ed_c_dt, 1, 1 );
    sprintf(str,"%.1f",cycle_delay);
    ed_c_dt->setText( str );
    ed_c_dt->setMinimumSize( 30, 15 );
    ed_c_dt->setMaximumSize( 100, 30 );
    QToolTip::add(ed_c_dt, "Delay (sec) after adding" );
    lb = new QLabel( w );
    g->addWidget( lb, 1, 0 );
    lb->setText( "Delay:" );
    lb->setBuddy( ed_c_dt );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Delay (sec) after adding" );

    cb_bg = new QCheckBox( w );
    g->addWidget( cb_bg, 3, 0 );
    cb_bg->setText("Subtr.BGD");
    cb_bg->setMinimumSize( 30, 15 );
    cb_bg->setMaximumSize( 100, 30 );
    connect( cb_bg, SIGNAL(clicked()), SLOT(toggle_bias()) );
    QToolTip::add(cb_bg, "Do Background subtruction" );
    cb_fl = new QCheckBox( w );
    g->addWidget( cb_fl, 3, 1 );
    cb_fl->setText("Div.FLAT");
    cb_fl->setMinimumSize( 30, 15 );
    cb_fl->setMaximumSize( 100, 30 );
    connect( cb_fl, SIGNAL(clicked()), SLOT(toggle_flat()) );
    QToolTip::add(cb_fl, "Divide by FLAT" );

    ed_c_n = new QLineEdit( w );
    g->addWidget( ed_c_n, 2, 1 );
    sprintf(str,"%d",cycle_n);
    ed_c_n->setText( str );
    ed_c_n->setMinimumSize( 30, 15 );
    ed_c_n->setMaximumSize( 100, 30 );
    QToolTip::add(ed_c_n, "Number of cycles (-1 - infinitely)" );
    lb = new QLabel( w );
    g->addWidget( lb, 2, 0 );
    lb->setText( "N cycles:" );
    lb->setBuddy( ed_c_n );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Number of cycles (-1 - infinitely)" );

    // finally insert page "cycle" into the tab dialog and start GM
    td_opt->addTab( w, "&Cycle" );
    g->activate();

    // set up 2-nd page  of the tab dialog
    w = new QWidget( td_opt, "cuts" );
    g = new QGridLayout( w, 2, 2, 5 );

    ed_bcut = new QLineEdit( w );
    g->addWidget( ed_bcut, 0, 1 );
    sprintf(str,"%d",cutBot);
    ed_bcut->setText( str );
    ed_bcut->setMinimumSize( 30, 15 );
    ed_bcut->setMaximumSize( 100, 30 );
    QToolTip::add(ed_bcut, "Low intensity value" );
    lb = new QLabel( w );
    g->addWidget( lb, 0, 0 );
    lb->setText( "Bottom:" );
    lb->setBuddy( ed_bcut );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Low intensity value" );

    ed_tcut = new QLineEdit( w );
    g->addWidget( ed_tcut, 1, 1 );
    sprintf(str,"%d",cutTop);
    ed_tcut->setText( str );
    ed_tcut->setMinimumSize( 30, 15 );
    ed_tcut->setMaximumSize( 100, 30 );
    QToolTip::add(ed_tcut, "Hight intensity value" );
    lb = new QLabel( w );
    g->addWidget( lb, 1, 0 );
    lb->setText( "Top:" );
    lb->setBuddy( ed_tcut );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Hight intensity value" );

    // finally insert page "cuts" into the tab dialog and start GM
    td_opt->addTab( w, "C&UT's" );
    g->activate();

    // set up 3-d page  of the tab dialog
    w = new QWidget( td_opt, "frame" );
    g = new QGridLayout( w, 4, 2, 5 );
    tv_chan_get_frame(&frame_x, &frame_y, &frame_w, &frame_h);

    ed_f_x = new QLineEdit( w );
    g->addWidget( ed_f_x, 0, 1 );
    sprintf(str,"%d",frame_x);
    ed_f_x->setText( str );
    ed_f_x->setMinimumSize( 30, 15 );
    ed_f_x->setMaximumSize( 100, 30 );
    QToolTip::add(ed_f_x, "X-position of Upper-Left conner" );
    lb = new QLabel( w );
    g->addWidget( lb, 0, 0 );
    lb->setText( "FrameX:" );
    lb->setBuddy( ed_f_x );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "X-position of Upper-Left conner" );

    ed_f_y = new QLineEdit( w );
    g->addWidget( ed_f_y, 1, 1 );
    sprintf(str,"%d",frame_y);
    ed_f_y->setText( str );
    ed_f_y->setMinimumSize( 30, 15 );
    ed_f_y->setMaximumSize( 100, 30 );
    QToolTip::add(ed_f_y, "Y-position of Upper-Left conner" );
    lb = new QLabel( w );
    g->addWidget( lb, 1, 0 );
    lb->setText( "FrameY:" );
    lb->setBuddy( ed_f_y );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Y-position of Upper-Left conner" );

    ed_f_w = new QLineEdit( w );
    g->addWidget( ed_f_w, 2, 1 );
    sprintf(str,"%d",frame_w);
    ed_f_w->setText( str );
    ed_f_w->setMinimumSize( 30, 15 );
    ed_f_w->setMaximumSize( 100, 30 );
    QToolTip::add(ed_f_w, "Frame width");
    lb = new QLabel( w );
    g->addWidget( lb, 2, 0 );
    lb->setText( "Width:" );
    lb->setBuddy( ed_f_w );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Frame width" );

    ed_f_h = new QLineEdit( w );
    g->addWidget( ed_f_h, 3, 1 );
    sprintf(str,"%d",frame_h);
    ed_f_h->setText( str );
    ed_f_h->setMinimumSize( 30, 15 );
    ed_f_h->setMaximumSize( 100, 30 );
    QToolTip::add(ed_f_h, "Frame height");
    lb = new QLabel( w );
    g->addWidget( lb, 3, 0 );
    lb->setText( "Height:" );
    lb->setBuddy( ed_f_h );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Frame height" );

    // finally insert page "frame" into the tab dialog and start GM
    td_opt->addTab( w, "&Frame" );
    g->activate();

    // set up 4-th page  of the tab dialog
    w = new QWidget( td_opt, "vs56" );
    g = new QGridLayout( w, 2, 2, 5 );

    ed_exp = new QLineEdit( w );
    g->addWidget( ed_exp, 0, 1 );
    sprintf(str,"%.2f", tv_chan_exp(-1.));
    ed_exp->setText( str );
    ed_exp->setMinimumSize( 30, 15 );
    ed_exp->setMaximumSize( 100, 30 );
    QToolTip::add(ed_exp, "Tandem exposure (sec)" );
    lb = new QLabel( w );
    g->addWidget( lb, 0, 0 );
    lb->setText( "Exposure:" );
    lb->setBuddy( ed_exp );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Tandem exposure (sec)" );

    ed_elref = new QLineEdit( w );
    g->addWidget( ed_elref, 1, 1 );
    sprintf(str,"%d",tv_chan_elref(-1));
    ed_elref->setText( str );
    ed_elref->setMinimumSize( 30, 15 );
    ed_elref->setMaximumSize( 100, 30 );
    QToolTip::add(ed_elref, "Electronic reference level of VS56 ADC" );
    lb = new QLabel( w );
    g->addWidget( lb, 1, 0 );
    lb->setText( "El.ref.level:" );
    lb->setBuddy( ed_elref );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Electronic reference level of VS56 ADC" );

    // finally insert page "vs56" into the tab dialog and start GM
    td_opt->addTab( w, "VS56" );
    g->activate();



    // we want Apply & Cancel in tab.dialog
    td_opt->setApplyButton();
    connect( td_opt, SIGNAL(applyButtonPressed()), SLOT(apply_opt()) );
    td_opt->setCancelButton();

    td_opt->show();

}
void TVimage::destr_opt()  // fix that "Options" dialog window was destroyed
{
    td_opt = NULL;
}
void TVimage::apply_opt()  // Use new (edited) values from "Options" dialog
{
    cycle_nadd  = atoi( ed_c_nadd->text() );
    if (cycle_nadd<1) cycle_nadd=1;
    cycle_delay = atof( ed_c_dt->text() );
    if (cycle_delay<0.1) cycle_delay=0.1;
    cycle_n     = atoi( ed_c_n->text() );

    int cb = cutBot;
    cutBot = atoi( ed_bcut->text() );
    if (cutBot<-32768) cutBot=-32768;
    int ct = cutTop;
    cutTop = atoi( ed_tcut->text() );
    if (cutTop>32767) cutTop=32767;
    if (cutTop<cutBot+10) cutTop=cutBot+10;
    if (cb!=cutBot || ct!=cutTop)
       show_image(TRUE);

    int fx = frame_x;
    frame_x = atoi( ed_f_x->text() );
    int fy = frame_y;
    frame_y = atoi( ed_f_y->text() );
    int fw = frame_w;
    frame_w = atoi( ed_f_w->text() );
    int fh = frame_h;
    frame_h = atoi( ed_f_h->text() );
    if (fx != frame_x || fy != frame_y || fw != frame_w || fh != frame_h )
       tv_chan_frame(frame_x,frame_y,frame_w,frame_h);

    double exp = atof( ed_exp->text() );
    if (exp<0.04) exp=0.04;
    tv_chan_exp(exp);
    int elref = atoi( ed_elref->text() );
    if (elref<0) elref=0;
    if (elref>63) elref=63;
    if (elref != tv_chan_elref(-1))
       tv_chan_elref(elref);
}

void TVimage::show_text( const char *fileName )
{
    QFile f( fileName );
    if ( !f.open( IO_ReadOnly ) )
	return;

    QWidget *tw = new QWidget(0,"Show Text Window");
    QMultiLineEdit *text = new QMultiLineEdit( tw, "Text" );


    text->setReadOnly ( TRUE );
    text->setAutoUpdate( FALSE );
    text->clear();

    QTextStream t(&f);
    while ( !t.eof() ) {
	QString s = t.readLine();
	text->append( s );
    }
    f.close();
    text->setAutoUpdate( TRUE );
    text->resize(500,500);
    text->repaint();
    tw->setCaption( fileName );
    tw->resize(500,500);
    tw->show();
}
void TVimage::show_help() {
    show_text("Doc/VS56.txt");
}
void TVimage::html_manual() {
    system("netscape Doc/vs56.html &");
}

void TVimage::single()
{
    if(cycle_stage!=0)  stop_cycle();
    if(vs56_ok())
       tv_chan_go(1,1,1,ccd_mode);
    else
       mesg->setText("No VS56 driver!!!");
}
void TVimage::get10()
{
    if(cycle_stage!=0)  stop_cycle();
    if(vs56_ok())
       tv_chan_go(1,10,1,ccd_mode);
    else
       mesg->setText("No VS56 driver!!!");
}
void TVimage::cont()
{
    if(cycle_stage!=0)  stop_cycle();
    if(vs56_ok())
       tv_chan_go(1,-1,1,ccd_mode);
    else
       mesg->setText("No VS56 driver!!!");
}
void TVimage::stop()
{
    if(cycle_stage!=0)  stop_cycle();
    if(vs56_ok()) {
       tv_chan_stop();
    } else
       mesg->setText("No VS56 driver!!!");
}
void TVimage::clear()
{
    if(cycle_stage!=0)  stop_cycle();
    if(vs56_ok())
       tv_chan_clear();
    else
       mesg->setText("No VS56 driver!!!");
}
void TVimage::init()
{
    if(cycle_stage!=0)  stop_cycle();
    if(vs56_ok())
       tv_chan_init();
    else
       mesg->setText("No VS56 driver!!!");
}
void TVimage::write()
{
    if(vs56_ok())
       tv_chan_write();
    else
       mesg->setText("No VS56 driver!!!");
}
void TVimage::add1()
{
    if(cycle_stage!=0)  stop_cycle();
    if(vs56_ok())
       tv_chan_go(0,1,1,ccd_mode);
    else
       mesg->setText("No VS56 driver!!!");
}
void TVimage::add10()
{
    if(cycle_stage!=0)  stop_cycle();
    if(vs56_ok())
       tv_chan_go(0,1,10,ccd_mode);
    else
       mesg->setText("No VS56 driver!!!");
}
void TVimage::addcont()
{
    if(cycle_stage!=0)  stop_cycle();
    if(vs56_ok())
       tv_chan_go(0,1,-1,ccd_mode);
    else
       mesg->setText("No VS56 driver!!!");
}
void TVimage::start_cycle()
{
    int dt = 200;
//puts("start_cycle()");

    if(!vs56_ok()) {
       mesg->setText("No VS56 driver!!!");
       return;
    }
    if(cycle_stage!=0) {
       stop_cycle();
       dt = 1000;
    }
    tvtimer->stop();
    disconnect( tvtimer, SIGNAL(timeout()), this, 0);
    connect( tvtimer, SIGNAL(timeout()), this, SLOT(step_cycle()) );
    tvtimer->start( dt , TRUE);
}

void TVimage::toggle_bias() {
    cycle_bias = !cycle_bias;
}
void TVimage::toggle_flat() {
    cycle_flat = !cycle_flat;
}

void TVimage::stop_cycle()
{
//puts("stop_cycle()");
    cycle_stage=0;
    tv_chan_stop();
    tvtimer->stop();
    disconnect( tvtimer, SIGNAL(timeout()), this, 0);
    connect( tvtimer, SIGNAL(timeout()), this, SLOT(tmout()) );
    tvtimer->start( 500, TRUE);
}
void TVimage::step_cycle()
{
    static int omseq = 0;
    int mseq = tv_chan_seq_msg();
    int nseq = tv_chan_seq_num();
    if(mseq > omseq) {
       char *msg = tv_chan_msg();
       mesg->setText(msg);
       if (cycle_stage>1 && strncmp(msg,"Can't",5)==0 ) {
	  stop_cycle();
	  return;
       }
    }
    omseq = mseq;

    if(!vs56_ok()) {
       mesg->setText("No VS56 driver!!!");
       stop_cycle();
       return;
    }
    tvtimer->stop();

//printf("step_cycle(%d)\n",cycle_stage);
    switch(cycle_stage) {
       default:
       case 0:                          // start
	  cycle_cnt = cycle_n;
	  cycle_stage=1;
	  break;
       case 1:                          // start get&add
	  cycle_add_ok=nseq+cycle_nadd;
	  tv_chan_go(1,1,cycle_nadd,ccd_mode);
	  cycle_stage=2;
	  break;
       case 2:                          // wait for get&add; then show & delay
	  if (nseq>=cycle_add_ok) {
	     show_image(force_showing);
	     if(cycle_bias) subtr_bias();
	     if(cycle_flat) div_by_flat();
	     show_image(force_showing);
	     force_showing=FALSE;
	     cycle_stage=3;
	     if (cycle_n<0 || cycle_cnt>1) {
		tvtimer->start( (int)(cycle_delay*1000.), TRUE);
		return;
	     }
	  }
	  break;
       case 3:                          // delay end -> restart?
	  if (cycle_n<0 || cycle_cnt>1) {
	     if (cycle_cnt>1) cycle_cnt--;
	     cycle_stage=1;
	  } else {
	     stop_cycle();
	     return;
	  }
	  break;
    }
    tvtimer->start( 200, TRUE);
}

void TVimage::save_to_file(int file_format)
{
  int x,y;
  FILE *pd;
  char command[80];
  uint range = cutTop-cutBot;
  static QString sname;
  static QString sdir;
  static QFile f;
  int pos,ind;

  switch (file_format) {
     case 0:
     case 2:
	sname = QFileDialog::getSaveFileName(savedir, "*.jpg",this,savename);
	if ( sname.isEmpty() ) return;
	savename = sname.data();
	if(strstr(savename,".jpg"))
	   sprintf(command,"cjpeg -quality 90 >%s", savename);
	else
	   sprintf(command,"cjpeg -quality 90 >%s.jpg", savename);
	break;
     case 1:
     case 3:
	sname = QFileDialog::getSaveFileName(savedir, "*.gif",this,savename);
	if ( sname.isEmpty() ) return;
	savename = sname.data();
	if(strstr(savename,".gif"))
	  sprintf(command,"cjpeg -quality 100 | djpeg -gif >%s", savename);
	else
	  sprintf(command,"cjpeg -quality 100 | djpeg -gif >%s.gif", savename);
	break;
     case 4:
	sname = QFileDialog::getSaveFileName(savedir, "*.mt",this,savename);
	if ( sname.isEmpty() ) return;
	savename = sname.data();
	if(strstr(savename,".mt"))
	  sprintf(command,"%s", savename);
	else
	  sprintf(command,"%s.mt", savename);
	break;
  }
  sdir = sname;
  ind = sdir.length();
  pos = sdir.findRev('/',ind-1);
  sdir.truncate(pos);
  savedir = sdir.data();

  if (file_format<4) {
     if((pd = popen(command,"w")) == NULL) return;

     switch (file_format) {
	case 0:
	case 1:
	   fprintf(pd,"P5\n%d %d\n255\n",image_w,image_h);  /*PGMRAW*/
	   break;
	case 2:
	case 3:
	   fprintf(pd,"P6\n%d %d\n255\n",image_w,image_h);  /*PPMRAW*/
	   break;
     }
     for ( y=0; y<image_h; y++ ) {                     // scan image pixels
	 for ( x=0; x<image_w; x++ ) {
	     int pix = tv_image[y*image_w+x] - cutBot;
	     if(pix<0) pix=0;
	     pix = pix * 256 / range;
	     if(pix>255) pix=255;
	     switch (file_format) {
		case 0:
		case 1:
		  fputc(itt[pix],pd);
		  break;
		case 2:
		case 3:
		   QRgb cpix = lut[itt[pix]];
		   fputc(qRed(cpix),pd);
		   fputc(qGreen(cpix),pd);
		   fputc(qBlue(cpix),pd);
		   break;
	     }
	 }
     }
     pclose(pd);
  } else if(file_format==4) {
      char tmp[2880];
      long int clock = tv_chan_tstart();
      struct tm *tm = gmtime(&clock);
      int tstart = (tm->tm_hour)*3600 + (tm->tm_min)*60 + tm->tm_sec;
      int tend;
      int nadd    = tv_chan_added();
      int tv_elref= tv_chan_elref(-1);
      int tandem  = tv_chan_tandem();
      int tv_lev0 = tv_chan_addlev(); /*zero level already substructed from ADDed images*/
      int img_len = image_w*image_h;
      int i,mn,mx;
      double exp;

      clock = tv_chan_tcurr();
      tm = gmtime(&clock);
      tend = (tm->tm_hour)*3600 + (tm->tm_min)*60 + tm->tm_sec;

      if (tandem) {
	 exp = tv_chan_exp(-1.);
	 tend += (int)(exp+0.5);
      } else
	 exp = nadd * 0.04;

      f.setName( command );
      if ( !f.open( IO_WriteOnly ) ) {
	 sprintf(tmp, "Can't write FITS-format to %s", command);
	 mesg->setText(tmp);
	 return;
      }
      char *p = strrchr(command, '.');
      strcpy(p,".bdf");
      p = strrchr(command, '/') + 1;

      for(i=0, mn=mx=tv_image[0]; i<img_len; i++) {
	 int j = tv_image[i];
	 if(mn>j) mn = j;
	 if(mx<j) mx = j;
      }

      sprintf(tmp,"SIMPLE  =                    T / Standard FITS format                           ");
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"BITPIX  =                   16 / No. of bits per pixel                          ");
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"NAXIS   =                    2 / No. of axes in image                           ");
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"NAXIS1  =                  %-3d / No. of pixels                                  ",image_w);
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"NAXIS2  =                  %-3d / No. of pixels                                  ",image_h);
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"BSCALE  =                   1. / Scaling factor: r = f*i + z                    ");
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"BZERO   =                   0. / Zero offset: r = f*i + z                       ");
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"DATAMAX =              %6d. / Maximum data value                             ",mx);
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"DATAMIN =              %6d. / Minimum data value                             ",mn);
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"ORIGIN  = 'SAO-TVimage'        / Written by TVimage tool at BTA SAO RAS         ");
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"DATE    = '%02d/%02d/%02d'           / Date of writting: DD/MM/YY                     ",tm->tm_mday, tm->tm_mon+1, tm->tm_year );
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"FILENAME= '%-8s'           / Original file base-name                        ",p );
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"MIDASFTP= 'IMAGE   '           / MIDAS File Type                                ");
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"AUTHOR  = 'Shergin V.S.'       / MIDAS desc.: AUTHOR(1)                         ");
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"TELESCOP= '6m      '           / MIDAS desc.: TELESCOP(1)                       ");
      f.writeBlock ( tmp,80 );
      if (tandem)
      sprintf(tmp,"INSTRUME= 'Tandem TV-camera'   / MIDAS desc.: INSTRUME(1)                       ");
      else
      sprintf(tmp,"INSTRUME= 'TV-cannel %1d'        / MIDAS desc.: INSTRUME(1)                       ",nchan);
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"DATE-OBS= '%02d/%02d/%02d'           / MIDAS desc.: O_TIME(1)                         ",tm->tm_mday, tm->tm_mon+1, tm->tm_year );
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"TM-START=              %6d. / MIDAS desc.: O_TIME(5)                         ",tstart);
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"EXPTIME =              %7.2f / MIDAS desc.: O_TIME(7)                         ",exp);
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"TM-END  =              %6d. /                                                ",tend);
      f.writeBlock ( tmp,80 );
//    sprintf(tmp,"HISTORY  TV-BZERO=     %6d. / zero level already subst-ed from image         ",tv_lev0);
      sprintf(tmp,"TV-BZERO=              %6d. / zero level already subst-ed from image         ",tv_lev0);
      f.writeBlock ( tmp,80 );
//    sprintf(tmp,"HISTORY  TV-ELREF=     %6d. / electronic reference level of graber's ADC     ",tv_elref);
      sprintf(tmp,"TV-ELREF=              %6d. / electronic reference level of graber's ADC     ",tv_elref);
      f.writeBlock ( tmp,80 );
//    sprintf(tmp,"HISTORY  TV-NADD =     %6d. / number of input images summarized              ",nadd);
      sprintf(tmp,"TV-NADD =              %6d. / number of input images summarized              ",nadd);
      f.writeBlock ( tmp,80 );
      sprintf(tmp,"END                                                                             ");
      f.writeBlock ( tmp,80 );

      int last_blk = ((img_len-1)/1440+1)*1440;
      f.at(2880);
      for (int blk=0; blk<last_blk; blk += 1440) {
	  for(i=0; i<1440; i++) {
	     int j=blk+i;
	     if(j<img_len) {
		union {char b[2];short i;} pix;
		pix.i = (int)tv_image[j];
		tmp[i*2+1] = pix.b[0];
		tmp[i*2]   = pix.b[1];
	     } else
		tmp[i*2] = tmp[i*2+1] = 0;
	  }
	  f.writeBlock ( tmp,2880 );
      }
      f.close();
  }
}
void TVimage::save_fits()
{
   save_to_file(4);
}

void TVimage::set_bias_name()
{
  static char fitsname[80];
  static char tmp[100];
  static QString fname;

  fname = QFileDialog::getOpenFileName(servdir, "*.mt*",this,biasname);
  if ( fname.isEmpty() ) return;

  if(strstr(fname.data(),".mt"))
    sprintf(fitsname,"%s", fname.data());
  else
    sprintf(fitsname,"%s.mt", fname.data());

  fname.truncate( fname.findRev( '/', fname.length()-1 ) );
  servdir = fname.data();

  biasname = fitsname;
  sprintf(tmp, "BGD-name now: %s", biasname);
  mesg->setText(tmp);
}
void TVimage::set_flat_name()
{
  static char fitsname[80];
  static char tmp[100];
  static QString fname;

  fname = QFileDialog::getOpenFileName(servdir, "*.mt*",this,flatname);
  if ( fname.isEmpty() ) return;

  if(strstr(fname.data(),".mt"))
    sprintf(fitsname,"%s", fname.data());
  else
    sprintf(fitsname,"%s.mt", fname.data());

  fname.truncate( fname.findRev( '/', fname.length()-1 ) );
  servdir = fname.data();

  flatname = fitsname;
  sprintf(tmp, "FLAT-name now: %s", flatname);
  mesg->setText(tmp);
}

void TVimage::bias_flat_red() {
   subtr_bias();
   div_by_flat();
}

void TVimage::subtr_bias() {
  int x,y,i;
  int sx, sy, sw, sh, xe, ye;
  int bs_lev0 = 0;
  int bs_nadd = 1;
  int bs_elref= 60;
  int tv_nadd = tv_chan_added();
//int tv_elref= tv_chan_elref(-1);
  int tv_lev0 = tv_chan_addlev();
  double s,k;

  if(biasname==NULL) return;

  short *bias = (short *)calloc(maxL, sizeof(short));
  if(bias==NULL) return;

  tv_chan_get_frame(&sx, &sy, &sw, &sh);
  xe = sx+sw;
  ye = sy+sh;

  if (!get_fits(biasname, bias, &bs_lev0, &bs_elref, &bs_nadd)) {
     free(bias);
     return;
  }
//printf("tv_nadd=%d bs_nadd=%d\n",tv_nadd,bs_nadd);
//printf("tv_lev0=%d bs_lev0=%d\n",tv_lev0,bs_lev0);
  k = (double)tv_nadd/(double)bs_nadd;

  for ( s=0.,y=0; y<image_h; y++ ) {
     for ( x=10; x<image_w-10; x++ ) {
	  if(x==100) x=image_w-100;
	  s += bias[y*image_w + x];
     }
  }
  s = s/(image_h*180.);
//printf("K=%f Bavr=%f\n",k,s);
  for ( y=sy; y<ye; y++ )
     for ( x=sx; x<xe; x++ ) {
	     i = y*image_w + x;
	     tv_image[i] -= (short)(((double)bias[i] - s)* k + 0.5);
     }
  tv_lev0 -= (int)((s+bs_lev0)*k+0.5);
  tv_chan_set_addlev(tv_lev0);
//printf("tv_lev0=%d\n",tv_lev0);
  free(bias);
  force_showing=TRUE;
}

void  TVimage::div_by_flat() {
  int x,y,i;
  int sx, sy, sw, sh, xe, ye;
  int ft_lev0 = 0;
  int ft_nadd = 1;
  int ft_elref= 60;
//int tv_nadd = tv_chan_added();
//int tv_elref= tv_chan_elref(-1);
  int tv_lev0 = tv_chan_addlev();
  double s;

  if(flatname==NULL) return;

  short *flat = (short *)calloc(maxL, sizeof(short));
  if(flat==NULL) return;

  tv_chan_get_frame(&sx, &sy, &sw, &sh);
  xe = sx+sw;
  ye = sy+sh;

  if (!get_fits(flatname, flat, &ft_lev0, &ft_elref, &ft_nadd)) {
     free(flat);
     return;
  }
  for ( s=0.,y=100; y<image_h-100; y++ )
     for ( x=100; x<image_w-100; x++ )
	  s += flat[y*image_w + x];
  s = s / ((image_h-200.)*(image_w-200.)) + ft_lev0;

  for ( y=sy; y<ye; y++ )
     for ( x=sx; x<xe; x++ ) {
	 i = y*image_w + x;
	 double f = flat[i]+ft_lev0;
	 if(f>-0.9&&f<0.9) f=1.;
	 f = s/f;
	 tv_image[i] = (short)((tv_image[i]+tv_lev0)*f-tv_lev0+0.5);
     }
  free(flat);
  force_showing=TRUE;
}

void TVimage::load_fits()
{
  char fitsname[80];
  static QString fname;
  static QString fdir;
  int tv_lev0 = 0;
  int nadd = 1;
  int tv_elref= 60;

  fname = QFileDialog::getOpenFileName(savedir, "*.mt*",this,savename);
  if ( fname.isEmpty() ) return;
  savename = fname.data();

  if(strstr(savename,".mt"))
    sprintf(fitsname,"%s", savename);
  else
    sprintf(fitsname,"%s.mt", savename);

  fdir = fname;
  fdir.truncate( fdir.findRev( '/', fdir.length()-1 ) );
  savedir = fdir.data();

  if (get_fits(fitsname, tv_image, &tv_lev0, &tv_elref, &nadd)) {
     tv_chan_set_addlev(tv_lev0);
     tv_chan_set_added(nadd);
     force_showing=TRUE;
  }
}

bool TVimage::get_fits(char *fitsname, short *array, int *tv_lev0, int *tv_elref, int *nadd)
{
  int x,y,lin;
  char tmp[2880];
  QFile f;
  bool simple = FALSE;
  int bitpix = 8;
  int naxis = 0;
  int xsize = 0;
  int ysize = 0;
  double bscale = 1.;
  double bzero = 0.;
  int mh, mn=32767, mx=-32768;
  bool pipe = FALSE;

  if(strstr(fitsname,".gz")) {
     char command[100];
     FILE *pd;
     sprintf(command,"zcat %s",fitsname);
     if((pd = popen(command,"r")) == NULL) {
	sprintf(tmp, "Can't unZIP FITS-file %s", fitsname);
	mesg->setText(tmp);
	return FALSE;
     }
     f.open(IO_ReadOnly, pd);
     pipe = TRUE;
  } else {
     f.setName( fitsname );
     if ( !f.open( IO_ReadOnly ) ) {
	sprintf(tmp, "Can't read FITS-file %s", fitsname);
	mesg->setText(tmp);
	return FALSE;
     }
     pipe = FALSE;
  }
  for (lin=36*5; lin>0; lin--) {
     if (f.readBlock ( tmp,80 ) <= 0 ) goto wrong_fits;
     QString pname(tmp,10);
     QString pval(tmp+9,70);
     int i = pval.find('/');
     if(i>0) pval.resize(i+1);
//printf("%s%s%d\n",(const char *)pname,(const char *)pval,i);

     if(       pname=="END      ") break;
     else if(  pname=="SIMPLE  =") {
	if( !(pval.find(" T ") || pval.find(" T/")) ) goto wrong_fits;
	simple = TRUE;
     } else if(pname=="BITPIX  =") {
	if( !(pval.find(" 16 ") || pval.find(" 16/"))) goto wrong_fits;
	bitpix = 16;
     } else if(pname=="NAXIS   =") {
	if( !(pval.find(" 2 ") || pval.find(" 2/"))) goto wrong_fits;
	naxis = 2;
     } else if(pname=="NAXIS1  =") {
	xsize = pval.toInt();
     } else if(pname=="NAXIS2  =") {
	ysize = pval.toInt();
     } else if(pname=="BSCALE  =") {
	bscale = pval.toDouble();
     } else if(pname=="BZERO   =") {
	bzero = pval.toDouble();
     } else if(pname=="TV-BZERO=") {
	*tv_lev0 = (int)pval.toDouble();
     } else if(pname=="TV-ELREF=") {
	*tv_elref = (int)pval.toDouble();
     } else if(pname=="TV-NADD =") {
	*nadd = (int)pval.toDouble();
     } else if(pname=="HISTORY  ") {
	QString hname(pval.data(),10);
	QString hval(pval.data()+9);
	if(hname=="TV-BZERO=")
	   *tv_lev0 = (int)hval.toDouble();
	if(hname=="TV-ELREF=")
	   *tv_elref = (int)hval.toDouble();
	if(hname=="TV-NADD =")
	   *nadd = (int)hval.toDouble();
     }
  }
//printf("%dlines simple=%s bitpix=%d naxis=%d\n",36*5-lin,simple?"T":"F",bitpix,naxis);
//printf("xsize=%d  ysize=%d bscale=%f bzero=%f\n",xsize,ysize,bscale,bzero);

  if(lin==0 || !simple || bitpix!=16 || naxis!=2 || xsize<1 || ysize<1) goto wrong_fits;

  if(pipe) {
     lin = lin%36;
     while(--lin>0)
	if (f.readBlock ( tmp,80 ) <= 0 ) goto wrong_fits;
  } else
     f.at(((f.at()-1)/2880+1)*2880);

  if(image_w!=xsize || image_h!=ysize )
       resetFrame();

  mh = maxL/image_w;
  if (ysize>mh) ysize=mh;
  if (ysize>image_h) image_h=ysize;

  for ( y=0; y<ysize; y++ ) {
     for ( x=0; x<xsize; x++ ) {
	union {char b[2]; short i;} pix;
	int i;
	if ((i=f.getch())<0)  goto wrong_fits;
	pix.b[1] = i;
	if ((i=f.getch())<0)  goto wrong_fits;
	pix.b[0] = i;
	if(x<image_w) {
	   i = array[y*image_w+x] = (short)(pix.i*bscale + bzero + 0.5);
	   if(mn>i) mn = i;
	   if(mx<i) mx = i;
	}
     }
  }
  f.close();

  sprintf(tmp, "Load FITS: width=%d height=%d min=%d max=%d",xsize,ysize,mn,mx);
  mesg->setText(tmp);
  return TRUE;

wrong_fits:
  sprintf(tmp, "Can't recognize FITS-format (%s)", fitsname);
  mesg->setText(tmp);
  f.close();
  return FALSE;
}

void TVimage::stat_image() {
   char tmp[100];
   int sx, sy, sw, sh;
   int i,j,x,y,pix;
   int n,nbot,nmid,ntop;
   int hist0,hbot,htop,med,mx,mn;
   double dsp,avr;
   int *hist = (int *)calloc(65536, sizeof(int));
   if(hist==NULL) return;

   tv_chan_get_frame(&sx, &sy, &sw, &sh);

   n = sw*sh;
   nbot = (int)(n*0.05 + 0.5);
   nmid = n/2;
   ntop = (int)(n*0.95 + 0.5);

   for (j=0; j<65536; j++) hist[j]=0;
   hist0 = -32768;
   mn=32767;
   mx=-32768;
   for (j=0; j<sh; j++) {
      y=sy+j;
      for (i=0; i<sw; i++) {
	 x=sx+i;
	 pix = tv_image[image_w*y+x];
	 hist[pix-hist0]++;
	 if(mn>pix) mn = pix;
	 if(mx<pix) mx = pix;
      }
   }
   hbot = 0;
   med = 32767;
   htop = 65535;
   for (i=j=0; j<65535; j++) {
      i += hist[j];
      if(i<=nbot) hbot = j;
      if(i<=nmid) med  = j;
      if(i<=ntop) htop = j;
   }
   med += hist0;
   for (avr=0.,i=0,j=hbot; j<htop; j++) {
      avr += hist[j]*j;
      i += hist[j];
   }
   avr /= i;
   for (dsp=0.,j=hbot; j<htop; j++) {
      double d = j-avr;
      dsp += d*d*hist[j];
   }
   avr += hist0;
   dsp = sqrt(dsp/i);
   //cutBot = (int)(avr-3.*dsp+0.5);
   //cutTop = (int)(avr+5.*dsp+0.5);
   cutBot = hbot+hist0-(int)(dsp+0.5);
   cutTop = htop+hist0+(int)(dsp+0.5);
   sprintf(tmp,"New CUTs=%d,%d Min=%d Max=%d Med=%d Avr=%.1f Dsp=%.2f",cutBot,cutTop,mn,mx,med,avr,dsp);
   mesg->setText(tmp);

   if (td_opt != NULL) {
      sprintf(tmp,"%d",cutBot);
      ed_bcut->setText( tmp );
      sprintf(tmp,"%d",cutTop);
      ed_tcut->setText( tmp );
   }

   force_showing=TRUE;

   free(hist);
}

void TVimage::togle_ccd_mode() {
    set_ccd_mode(!ccd_mode);
    tv_chan_go(0,0,0,ccd_mode); // fix it in vs56-driver for other programms
}
void TVimage::set_ccd_mode(bool new_mode) {
    if(new_mode==ccd_mode) return;
    ccd_mode = new_mode;
    if(ccd_mode) {
      ccd->changeItem ( "CCD ->vid (Don't use &BgdRow)",  ccd_id );
    } else {
      ccd->changeItem ( "Vid ->ccd (Use &BgdRaw online)",  ccd_id );
    }
    ccd->setItemChecked ( ccd_id, ccd_mode );
}
void TVimage::calc_flat_row() {
    int x,y;
    int avr[maxW];
    int tv_lev0 = tv_chan_addlev(); /*zero level already substructed from ADDed images*/
    int nadd = tv_chan_added();

    for ( x=0; x<image_w; x++ ) avr[x] = 0;
    for ( y=1; y<image_h; y++ )                      // scan image rows
	for ( x=0; x<image_w; x++ )                  // scan row pixels
	     avr[x] += tv_image[y*image_w+x];  //     ...
    for ( x=0; x<image_w; x++ ) {
	tv_image[x] = avr[x]/image_h;
	avr[x] = ((avr[x]<<8)/image_h + (tv_lev0<<8))/nadd; //.  nadd=256
    }
    tv_chan_ccd_bias(0);
    tv_chan_ccd_flat(avr, 0,  image_w);

    line_x1=image_w-1;
    line_x0=0;
    line_y1=line_y0=0;
    draw_graph();
    sleep(2);
}

void TVimage::use_flat_row(int oper) {
    int x,y;
    int sx, sy, sw, sh, xe, ye;
    int favr=0;
//  int iavr=0;
    double s;
//  short *p;
    int *frow = tv_chan_ccd_flat(NULL, 0, 0);
    int flat[maxW];
    int tv_lev0 = tv_chan_addlev();
    int nadd = tv_chan_added();

    tv_chan_get_frame(&sx, &sy, &sw, &sh);
    xe = sx+sw;
    ye = sy+sh;

    if(oper==0 || oper==1) {
	//for ( s=0.,x=0; x<image_w; x++ ) {
	//    if(x==100) x=image_w-100;
	//    s += frow[x];
	//}
	//favr = (int)(s/200. + 0.5);
	//for ( p=tv_image,s=0.,y=0; y<image_h; y++ ) {
	//    for ( x=0; x<image_w; x++,p++ ) {
	//       if(x==100) x=image_w-100;
	//       s += tv_image[y*image_w+x];
	//    }
	//}
	//iavr = (int)(s/(200.*image_h) + tv_lev0 + 0.5);
	//s = (double)iavr / (double)favr;
//printf("ImageLev=%d\n",iavr);
	s = (double)nadd / 256.;
	for ( x=0; x<image_w; x++ )
	    flat[x] = (int)((double)frow[x] * s + 0.5);
	for ( s=0.,x=0; x<image_w; x++ )
	       s += flat[x];
	favr = (int)(s/image_w + 0.5);
    } else {
	for ( x=0; x<image_w; x++ )
	    flat[x] = (frow[x]==0 && oper==3)? 1 : frow[x];
	for ( s=0.,x=0; x<image_w; x++ )
	       s += frow[x];
	favr = (int)(s/image_w + 0.5);
	if(favr==0 && oper==2) favr=1;
    }
//printf("FlatLev=%d\n",favr);
    switch (oper) {
	case 0: //Add
	       for ( y=sy; y<ye; y++ )
		   for ( x=sx; x<xe; x++ ) {
			short *p = &tv_image[y*image_w+x];
			*p = (int)*p + flat[x] - favr;
		   }
	       break;
	case 1: //Sub
	       for ( y=sy; y<ye; y++ )
		   for ( x=sx; x<xe; x++ ) {
			short *p = &tv_image[y*image_w+x];
			*p = (int)*p - flat[x] + favr;
		   }
	       break;
	case 2: //Mul
	       for ( y=sy; y<ye; y++ )
		   for ( x=sx; x<xe; x++ ) {
			short *p = &tv_image[y*image_w+x];
			*p = ((int)*p+tv_lev0) * flat[x]/favr - tv_lev0;
		   }
	       break;
	case 3: //Div
	       for ( y=sy; y<ye; y++ )
		   for ( x=sx; x<xe; x++ ) {
			short *p = &tv_image[y*image_w+x];
			*p = ((int)*p+tv_lev0) * favr / flat[x] - tv_lev0;
		   }
	       break;
	default:
	       break;
    }
    force_showing=TRUE;
}

void TVimage::tmout()
{
//puts("tmout()");
    tvtimer->stop();
// int s = QTime::currentTime().second();
// int ms = QTime::currentTime().msec();
// printf("t=%d.%d  ",s,ms);
// fflush(stdout);
    show_image(force_showing);
    force_showing=FALSE;
    tvtimer->start( 200, TRUE );                // emit signal every 0.2 seconds
}

void TVimage::togleFormat()
{
static int f1w = 768;
static int f1h = 512;
static int f2w = 512;
static int f2h = 512;
    char tmp[80];
    if(tv_chan_form(-1)==1) {
       f1w = tv_chan_img_width();
       f1h = tv_chan_img_height();
    } else {
       f2w = tv_chan_img_width();
       f2h = tv_chan_img_height();
    }
    if(format==1) {
       format=2;
       sprintf(tmp,"Setting 2-nd format [%d:%d].",f2w,f2h);
    } else if(format==2) {
       format=1;
       sprintf(tmp,"Setting 1-st format [%d:%d].",f1w,f1h);
    }
    tv_chan_form(format);
    mesg->setText(tmp);
}

void TVimage::resetFrame()
{
    char tmp[80];
    frame_x = 0;
    frame_y = 0;
    frame_w = tv_chan_img_width();
    frame_h = tv_chan_img_height();
    tv_chan_frame(frame_x, frame_y, frame_w, frame_h);
    sprintf(tmp,"Reset Frame: [%d,%d:%d,%d]",
	       frame_x, frame_y, frame_x+frame_w, frame_y+frame_h);
    mesg->setText(tmp);
    if (td_opt != NULL) {
       sprintf(tmp,"%d",frame_x);
       ed_f_x->setText( tmp );
       sprintf(tmp,"%d",frame_y);
       ed_f_y->setText( tmp );
       sprintf(tmp,"%d",frame_w);
       ed_f_w->setText( tmp );
       sprintf(tmp,"%d",frame_h);
       ed_f_h->setText( tmp );
    }
    copy_image();
}

void TVimage::paintEvent( QPaintEvent * )
{
//  printf("\npaintEvent()\n");
    copy_image();
}

void TVimage::setMouseMode(int mm)
{
   mouse_mode = mm;
   switch(mm) {
      case MOD_CUTS:  setCursor(sizeHorCursor);   break;
      case SET_FRAME: setCursor(sizeFDiagCursor); break;
      case SET_LINE:  setCursor(sizeAllCursor);   break;
      case SET_POINT: setCursor(crossCursor);     break;
      default:        setCursor(arrowCursor);     break;
   }
}
void TVimage::mousePressEvent( QMouseEvent *e )
{
    char tmp[80];
    mouse_button = e->button();
    mouse_x0 = mouse_x = e->x();
    mouse_y0 = mouse_y = e->y();
    if (mouse_mode == MOD_CUTS) {
       cutBot0 = cutBot;
       cutTop0 = cutTop;
    }
    else if (mouse_mode == SET_FRAME) {
       frame_x=mouse_x0; /*frame_w=10;*/
       frame_y=mouse_y0; /*frame_h=10;*/
       lin.setRasterOp (XorROP);
       lin.drawRect(frame_x,frame_y,frame_w,frame_h);
       sprintf(tmp,"New Frame: [%d,%d:%d,%d]",
	       frame_x, frame_y-image_line0,
	       frame_x+frame_w, frame_y-image_line0+frame_h);
       mesg->setText(tmp);
       if (td_opt != NULL) {
	  sprintf(tmp,"%d",frame_x);
	  ed_f_x->setText( tmp );
	  sprintf(tmp,"%d",frame_y-image_line0);
	  ed_f_y->setText( tmp );
	  sprintf(tmp,"%d",frame_w);
	  ed_f_w->setText( tmp );
	  sprintf(tmp,"%d",frame_h);
	  ed_f_h->setText( tmp );
       }
    }
    else if (mouse_mode == SET_LINE) {
       line_x1=line_x0=mouse_x0;
       line_y1=line_y0=mouse_y0;
       lin.setRasterOp (XorROP);
    }
    else if (mouse_mode == SET_POINT) {
       point_x = mouse_x;
       point_y = mouse_y - image_line0;
       point_z = tv_image[image_w*point_y+point_x];
       sprintf(tmp,"Location: %d-%d   Intensity: %d",point_x,point_y,point_z);
       mesg->setText(tmp);
    }
}
void TVimage::mouseReleaseEvent( QMouseEvent *e )
{
    if(mouse_button == NoButton) return;
    mouse_button = NoButton;
    if (mouse_mode == MOD_CUTS) {
       force_showing=TRUE;
    }
    else if (mouse_mode == SET_FRAME) {
       frame_y -= image_line0;
       tv_chan_frame(frame_x,frame_y,frame_w,frame_h);
    }
    else if (mouse_mode == SET_LINE) {
       copy_image();
       lin.setRasterOp (CopyROP);
       line_y0 -= image_line0;
       line_y1 -= image_line0;
       draw_graph();
    }
    else if (mouse_mode == SET_POINT) {
       copy_image();
       lin.setRasterOp (CopyROP);
       QColor c;
       c.setNamedColor("Black");
       lin.setPen(c);
       lin.drawLine(mouse_x-6,mouse_y-4,mouse_x+4,mouse_y+6);
       lin.drawLine(mouse_x-4,mouse_y-6,mouse_x+6,mouse_y+4);
       lin.drawLine(mouse_x-6,mouse_y+4,mouse_x+4,mouse_y-6);
       lin.drawLine(mouse_x-4,mouse_y+6,mouse_x+6,mouse_y-4);
       c.setNamedColor("White");
       lin.setPen(c);
       lin.drawLine(mouse_x-5,mouse_y-5,mouse_x+5,mouse_y+5);
       lin.drawLine(mouse_x-5,mouse_y+5,mouse_x+5,mouse_y-5);
    }
}
void TVimage::mouseMoveEvent( QMouseEvent *e )
{
    char tmp[80];
    if (mouse_button == NoButton) return;
    mouse_x = e->x();
    mouse_y = e->y();
    if (mouse_mode == MOD_CUTS) {
       int shft = mouse_x-mouse_x0;
       int dir = (shft>=0)? 1 : -1;
       int shft2 = dir*(shft*shft+32)/64;
       if (mouse_button == LeftButton || mouse_button == MidButton) {
	  cutBot = cutBot0 + shft2;
	  if (cutBot<-32768) cutBot=-32768;
	  if (cutBot>cutTop-10) cutBot=cutTop-10;
       }
       if (mouse_button == RightButton || mouse_button == MidButton) {
	  cutTop = cutTop0 + shft2;
	  if (cutTop>32767) cutTop=32767;
	  if (cutTop<cutBot+10) cutTop=cutBot+10;

       }
       sprintf(tmp,"New CUTs: %d,%d",cutBot,cutTop);
       mesg->setText(tmp);
       if (td_opt != NULL) {
	  sprintf(tmp,"%d",cutBot);
	  ed_bcut->setText( tmp );
	  sprintf(tmp,"%d",cutTop);
	  ed_tcut->setText( tmp );
       }
    }
    else if (mouse_mode == SET_FRAME) {
       if(abs(mouse_x-mouse_x0)<5 || abs(mouse_y-mouse_y0)<5 )
	 return;
       lin.drawRect(frame_x,frame_y,frame_w,frame_h);
       frame_x=mouse_x0; frame_w=mouse_x-mouse_x0;
       frame_y=mouse_y0; frame_h=mouse_y-mouse_y0;
       if(frame_w<0) {
	  frame_x += frame_w; frame_w = -frame_w;
       }
       if(frame_h<0) {
	  frame_y += frame_h; frame_h = -frame_h;
       }
       lin.drawRect(frame_x,frame_y,frame_w,frame_h);
       sprintf(tmp,"New Frame: [%d,%d:%d,%d]",
	       frame_x, frame_y-image_line0,
	       frame_x+frame_w, frame_y-image_line0+frame_h);
       mesg->setText(tmp);
       if (td_opt != NULL) {
	  sprintf(tmp,"%d",frame_x);
	  ed_f_x->setText( tmp );
	  sprintf(tmp,"%d",frame_y-image_line0);
	  ed_f_y->setText( tmp );
	  sprintf(tmp,"%d",frame_w);
	  ed_f_w->setText( tmp );
	  sprintf(tmp,"%d",frame_h);
	  ed_f_h->setText( tmp );
       }
    }
    else if (mouse_mode == SET_LINE) {
       lin.drawLine(line_x0,line_y0,line_x1,line_y1);
       lin.drawLine(mouse_x0,mouse_y0,mouse_x,mouse_y);
       sprintf(tmp,"New Line: %d,%d-%d,%d",mouse_x0,mouse_y0,mouse_x,mouse_y);
       mesg->setText(tmp);
       line_x0=mouse_x0; line_x1=mouse_x;
       line_y0=mouse_y0; line_y1=mouse_y;
    }
    else if (mouse_mode == SET_POINT) {
       point_x = mouse_x;
       point_y = mouse_y - image_line0;
       point_z = tv_image[image_w*point_y+point_x];
       sprintf(tmp,"Location: %d-%d   Intensity: %d",point_x,point_y,point_z);
       mesg->setText(tmp);
    }
}
void TVimage::getLUT()
{
   QString lname = QFileDialog::getOpenFileName("./lut", "*.lut",this, lut_name.data());
   if ( !lname.isEmpty() ) {
      getLUT( lname );
   }
}
void TVimage::getLUT(const char *lname)
{
   char tmp[80];
   QDir cwd = QDir::current();
   QDir::setCurrent("./lut");
   QFile f( lname );
   if ( !f.open( IO_ReadOnly ) ) {
	QDir::setCurrent(cwd.path());
	return;
   }
   lut_name = lname;
   sprintf(tmp,"Set LUT: %s",lname);
   mesg->setText(tmp);

   QTextStream t(&f);
   for (int i=0; !t.eof() && i<256; i++) {
       double r,g,b;
       sscanf(t.readLine(), "%lf %lf %lf", &r, &g, &b);
       lut[i] = qRgb((int)(r*255.), (int)(g*255.), (int)(b*255.));
   }
   force_showing=TRUE;
   f.close();
   QDir::setCurrent(cwd.path());
}
void TVimage::getITT()
{
   QString iname = QFileDialog::getOpenFileName("./itt", "*.itt",this, itt_name.data());
   if ( !iname.isEmpty() ) {
      getITT( iname );
   }
}
void TVimage::getITT(const char *iname)
{
   char tmp[80];
   QDir cwd = QDir::current();
   QDir::setCurrent("./itt");
   QFile f( iname );
   if ( !f.open( IO_ReadOnly ) ) {
	QDir::setCurrent(cwd.path());
	return;
   }
   itt_name =iname;
   sprintf(tmp,"Set ITT: %s",iname);
   mesg->setText(tmp);

   QTextStream t(&f);
   for (int i=0; !t.eof() && i<256; i++) {
       double v;
       sscanf(t.readLine(), "%lf", &v);
       itt[i] = (int)(v*255.);
   }
   force_showing=TRUE;
   f.close();
   QDir::setCurrent(cwd.path());
}
void TVimage::startExp()
{
    int n = (int)(sqrt(tv_chan_exp(-1.))*10.);
    if(tv_chan_tandem()) {
       expos = new QSlider(2, 300, 1, n, QSlider::Horizontal, this, "Exposure" );
       //expos->setTickmarks(QSlider::Both);
       //expos->setTickInterval(10);
       expos->setGeometry( 10, image_line0, width()-20, expos->height() );
       connect( expos, SIGNAL(valueChanged(int)), this, SLOT(getExp(int)) );
       connect( expos, SIGNAL(sliderReleased()), this, SLOT(stopExp()) );
       QToolTip::add( expos, "Set Exposure" );
       expos->show();
    } else {
       char tmp[80];
       sprintf(tmp,"Can't set Exposure for this TV-channel! (For Tandem only)");
       mesg->setText(tmp);
    }
}
void TVimage::getExp(int nexp)
{
    char tmp[80];
    double v = (double)nexp*nexp/100.;
    v = (double)((int)(v/0.04+0.01))*0.04;
    if(vs56_ok())
       tv_chan_exp(v);
    else {
       mesg->setText("No VS56 driver!!!");
       return;
    }
    int m = (int)(v/60.);
    if(m==0) {
       if(v<10)
	  sprintf(tmp,"set Exposure: %.2fs", v);
       else
	  sprintf(tmp,"set Exposure: %.1fs", v);
    }
    else
       sprintf(tmp,"set Exposure: %dm %ds", m, (int)(v-m*60));
    mesg->setText(tmp);
    if (td_opt != NULL) {
       sprintf(tmp,"%.2f", v);
       ed_exp->setText( tmp );
    }
}
void TVimage::stopExp()
{
    delete expos;
}
void TVimage::startElev()
{
    elev = new QSlider(0, 63, 1, tv_chan_elref(-1), QSlider::Horizontal, this, "El.Ref.Level" );
    elev->setTickmarks(QSlider::Both);
    elev->setTickInterval(1);
    elev->setGeometry( 10, height()-40, width()-40, elev->height()+20 );
    connect( elev, SIGNAL(valueChanged(int)), this, SLOT(getElev(int)) );
    connect( elev, SIGNAL(sliderReleased()), this, SLOT(stopElev()) );
    QToolTip::add( elev, "Electr.Ref.Level" );
    elev->show();
}
void TVimage::getElev(int lev)
{
    char tmp[80];
    int olev;

    if(vs56_ok()) {
       olev = tv_chan_elref(lev);
       sprintf(tmp,"Change El.Ref.Level: %d->%d",olev,lev);
       mesg->setText(tmp);
       if (td_opt != NULL) {
	  sprintf(tmp,"%d", lev);
	  ed_elref->setText( tmp );
       }
    } else
       mesg->setText("No VS56 driver!!!");
}
void TVimage::stopElev()
{
    delete elev;
}
void TVimage::draw_graph()
{
    int i,l,x,x0,y,y0,w,h,pix;
    double dx,dy,length;
    uint range = cutTop-cutBot;
    QColor cb,cw;
    cb.setNamedColor("Black");
    cw.setNamedColor("White");
    w = image_w;
    h = image_h;
    x = line_x1-line_x0;
    y = line_y1-line_y0;
    l = (int)((length = sqrt((double)x*x+y*y)) + 0.5);
    if(l==0) return;
    dx = (double)x/length;
    dy = (double)y/length;
    for(x0=y0=i=0;i<l;i++) {
       x = (int)(line_x0+dx*i+0.5);
       y = (int)(line_y0+dy*i+0.5);
       pix = tv_image[y*w+x] - cutBot;
       if(pix<0) pix=0;
       y = h+image_line0 - pix * h / range - 1;
       if(y>=h+image_line0) y=h+image_line0-1;
       if(y<image_line0) y=image_line0;
       x = i*w/l;
       lin.setPen(cb);
       if(i>0)
	 lin.drawLine(x0,y0,x0,y);
       lin.drawLine(x0,y,x,y);
       lin.setPen(cw);
       if(i>0)
	 lin.drawLine(x0+1,y0+1,x0+1,y+1);
       lin.drawLine(x0+1,y+1,x+1,y+1);
       x0 = x; y0 = y;
    }
}

void TVimage::show_image(bool force)
{
    int i,x,y,w,h;
    int sx, sy, sw, sh;
    static int oseq = -1;
    static int omseq = 0;
    uint range = cutTop-cutBot;

    if (!wtag) return;

    tv_image = tv_chan_image (nchan);

    int mseq = tv_chan_seq_msg();
    if(mseq > omseq) mesg->setText(tv_chan_msg());
    omseq = mseq;

    int nseq = tv_chan_seq_num();
    if(nseq == oseq && !force) return;
    oseq = nseq;

//puts("show_image()"); fflush(stdout);

    if(force) {
       w = image_w;
       h = image_h;
    } else {
       image_w = w = tv_chan_img_width();
       image_h = h = tv_chan_img_height();
    }
    if(show_bar) h += 10;

    wtag->resize(w, h + image_line0);
    image_pm.resize(image_w,image_h);

    tv_chan_get_frame(&sx, &sy, &sw, &sh);
    QImage frame( sw, sh, 8, 256 );             // create frame image
    QPixmap frame_pm( sw, sh);
    QImage bar( w, 10, 8, 256 );                // create color bar image
    bar_pm.resize( w, 10);

    for ( i=0; i<256; i++ ) {                   // build color table
	frame.setColor( i, lut[i] );
	bar.setColor( i, lut[i] );
    }
    for ( y=0; y<sh; y++ ) {                     // set frame image pixels
	uchar *p = frame.scanLine(y);
	for ( x=0; x<sw; x++ ) {
	    int pix = tv_image[(sy+y)*w + sx+x] - cutBot;
	    if(pix<0) pix=0;
	    pix = pix * 256 / range;
	    if(pix>255) pix=255;
	    pix = itt[pix];
	    if(pix>250) pix=250;  // some MAGIC!? I don't understand this....
	    *p++ = pix;
	}
    }
    if(sx==0 && sy==0 &&  image_w <= sw  &&  image_h <= sh) {
       image_pm = frame;                    // convert frame image to pixmap
    } else {
       frame_pm = frame;
       // copy frame pixmap to whole image pixmap
       bitBlt( &image_pm, sx, sy, &frame_pm, 0, 0, sw, sh);
    }
    for ( y=0; y<10; y++ ) {                     // set bar image pixels
	uchar *p = bar.scanLine(y);
	for ( x=0; x<w; x++ ) {
	    int pix = x * 256 / w;
	    if(pix>255) pix=255;
	    *p++ = itt[pix];
	}
    }
    bar_pm = bar;                    // convert color bar image to pixmap

    copy_image();
}

void TVimage::togleBar()
{
    show_bar = !show_bar;
    if(show_bar) {
       wtag->resize(image_w, image_h + image_line0 + 10);
       copy_image();
    } else {
       wtag->resize(image_w, image_h + image_line0);
    }
}

void TVimage::copy_image()
{
    int dx, dy;
//  printf("copy_image()\n");
    dy = menu->height() + mesg->height();
    dx = 0;
    bitBlt( wtag, dx, dy, &image_pm);     // copy pixmap to window
    if(show_bar) {
       // copy color bar pixmap to window
       bitBlt( wtag, 0, dy+image_h, &bar_pm, 0, 0, image_w, 10);
    }
}
//void TVimage::copy_image()
//{
//    int dx, dy, sx, sy, sw, sh;
//    tv_chan_get_frame(&sx, &sy, &sw, &sh);
//    dy = sy + menu->height() + mesg->height();
//    dx = sx;
//    bitBlt( wtag, dx, dy, &image_pm, sx, sy, sw, sh);// copy pixmap to window
//    if(show_bar) {
//       dy = image_h + menu->height() + mesg->height();
//       bitBlt( wtag, 0, dy, &image_pm, 0, image_h, image_w, 10);
//    }
//}

//
// The program starts here.
//

int main( int argc, char **argv )
{
    char cap[40];
    int nc=0;

    QApplication app( argc, argv );
    int ac=qApp->argc();
    char **av=qApp->argv();
    TVimage tvi;

    if ( ac > 1 &&  strncmp(av[1],"-r",2) == 0 ) {
       nc =(ac>2)? atoi(av[2]) : 0;
       tvi.setTagWnd(app.desktop());
    } else {
       nc =(ac>1)? atoi(av[1]) : 0;
       tvi.setTagWnd(&tvi);
    }
    tvi.setChan(nc);
    sprintf(cap,"TVimage Chan.%d", nc);
    tvi.setCaption(cap);
    tvi.set_ccd_mode(tv_chan_ccd_mode() != 0);
    app.setMainWidget( &tvi );
    tvi.show();

    return app.exec();
}
